import 'package:flutter/foundation.dart';
import 'package:flutter/widgets.dart';
import 'package:flutter_live_data/flutter_live_data.dart';
import 'controller.dart';
import 'vm.dart';
import 'commons/bloc.dart' as bloc;

/// View
abstract class View<VM extends ViewModel> extends StatelessWidget {
  View(VM viewModel) : this._viewModel = viewModel;

  final VM _viewModel;

  VM get $vm => _viewModel;

  String get name => '{{sys.ui.View}}';

  final _ChildrenController children = _ChildrenController();

//  final Map<Symbol, ControllerWidget> _controllers = {};
//
//  T getController<T extends ControllerWidget>([Symbol name]) =>
//      _controllers[name ?? Symbol('$T')] as T;
//
//  T getViewModel<T extends ViewModel>([Symbol name]) => getController(name)?.viewModel as T;

  Widget plug<T extends ViewModel>(Widget controller, [Symbol name]) {
    name ??= Symbol('$T');
    if (controller is ControllerWidget) {
      children.set(name, controller as ControllerWidget);
    } else
      throw TypeError();
    return controller;
  }

  @override
  Widget build(BuildContext context);

  Widget $<T>(
    LiveData<T> $vm, {
    Symbol id,
    @required Widget Function(BuildContext context, T value) builder,
  }) {
    return bloc.$watch($vm, id: id, builder: builder);
  }

  Widget $watch<T>(
    LiveData<T> $vm, {
    Symbol id,
    @required Widget Function(BuildContext context, T value) builder,
  }) {
    return bloc.$watch($vm, id: id, builder: builder);
  }

  Widget $bool<T extends bool>(
    LiveData<T> $vm, {
    Symbol id,
    bool Function(T) predicate,
    Widget Function(BuildContext context, T value) $true,
    Widget Function(BuildContext context, T value) $false,
  }) {
    return bloc.$bool($vm, id: id, predicate: predicate, $true: $true, $false: $false);
  }

  Widget $switch<T>(
    LiveData<T> $vm, {
    Symbol id,
    @required Map<T, Widget Function(BuildContext context, T value)> builders,
    Widget Function(BuildContext context, T value) $default,
  }) {
    return bloc.$switch($vm, id: id, builders: builders, $default: $default);
  }

  Widget $if<T>(
    LiveData<T> $vm, {
    Symbol id,
    bool Function(T) condition,
    @required Widget Function(BuildContext context, T value) builder,
    Widget Function(BuildContext context, T value) $else,
  }) {
    return bloc.$if($vm, id: id, condition: condition, $else: $else, builder: builder);
  }

  Widget $else<T>(
    LiveData<T> $vm, {
    Symbol id,
    bool Function(T) condition,
    @required Widget Function(BuildContext context, T value) builder,
  }) {
    return bloc.$else($vm, id: id, condition: condition, builder: builder);
  }
}

class _ChildrenController {
  final Map<Symbol, ControllerWidget> _controllers = {};

  void set(Symbol name, ControllerWidget controller) => _controllers[name] = controller;

  T widget<T extends ControllerWidget>([Symbol name]) => _controllers[name ?? Symbol('$T')] as T;

  T vm<T extends ViewModel>([Symbol name]) => widget(name)?.viewModel as T;
}
